(require '[clojure.core.reducers :as r])


; split in words
(defn get-words [text] (re-seq #"\w+" text))

; word frequencies

; reduce function used for updating the count with the addition of a word
(defn counting [counts word]
  (assoc counts
         word (inc (get counts word 0))))

(defn word-frequencies [words]
  (reduce
   counting
   {} words))

; Note: frequencies already defined
; (frequencies ...)


; merge maps with +
(def merge-counts (partial merge-with +))

; use frequencies and mapcat: map a function and catenate the results
(defn count-words-sequential [strings]
  (reduce merge-counts
          (map word-frequencies
               (map get-words strings))))

; count words parallel
(defn count-words-parallel [strings]
  (reduce merge-counts
	  (pmap #(word-frequencies (get-words %)) strings)))

; random string generator
(def VALID-CHARS
  (map char (concat ;(range 48 58) ; 0-9
                    ;(range 66 91) ; A-Z
                    ;(range 97 123)))) ; a-z
                     (range 97 104)))) ; a-d

(defn random-char []
  (nth VALID-CHARS (rand (count VALID-CHARS))))

; random word of at most 'length' chars (and at least 2)
(defn random-word [length]
  (apply str (take (rand 2 length) (repeatedly random-char))))
 
; n words at most 'length' long
(use '[clojure.string :only (join)])
(defn random-str [n length]
  (join " " (take n (repeatedly  #(apply str (take length (repeatedly random-char)))))))

; 100000 strings of 100 words, each at most 4 chars long
(def strings
     "a sequence of 100000 strings of 100 words, each at most 4 chars long"
     (take 100000 (repeatedly #(random-str 10 4))))

; check performances
;(time (def result (count-words-sequential/parallel[-batch]) strings)))

; sequential version with functions composed, avoid double
; traversal (useless, done automatically)
(defn count-words-sequential-comp [strings]
  (reduce merge-counts
          (map (comp word-frequencies get-words) strings)))

; batching in blocks of size n
(defn count-words-parallel-batch [strings n]
  (reduce merge-counts
          (pmap count-words-sequential            ;; try count word seq comp
                (partition-all n strings))))

; overall number of strings is 100
; (time (def result (count-words-parallel-batch strings 5-20)))

; using fold (worse ... again, too many procs)

(defn count-words-fold [strings]
  (r/fold merge-counts counting (mapcat get-words strings)))


;; (defn parallel-frequencies [coll]
;;   (r/fold 
;;    (partial merge-with +)
;;    (fn [counts x] (assoc counts x 
;;                          (inc (get counts x 0))))
;;    coll))

;; (defn count-words-fold [strings]
;;   (parallel-frequencies (mapcat get-words strings)))
